from __future__ import annotations

from typing import Iterable, Mapping

import pytest

from link.domain import events
from link.domain.custom_types import Identifier
from link.domain.state import Commands, Components, Operations, State, Transition, states
from link.service.gateway import LinkGateway
from link.service.uow import UnitOfWork
from tests.assignments import create_assignments, create_identifier, create_identifiers

from .gateway import FakeLinkGateway


def initialize(assignments: Mapping[Components, Iterable[str]]) -> tuple[FakeLinkGateway, UnitOfWork]:
    gateway = FakeLinkGateway(create_assignments(assignments))
    return gateway, UnitOfWork(gateway)


def get_entity_states(gateway: LinkGateway, identifiers: Iterable[Identifier]) -> set[tuple[Identifier, type[State]]]:
    entity_states = set()
    for identifier in identifiers:
        entity = gateway.create_entity(identifier)
        entity_states.add((entity.identifier, entity.state))
    return entity_states


def test_updates_are_applied_to_gateway_on_commit() -> None:
    gateway, uow = initialize({Components.SOURCE: {"1", "2"}, Components.OUTBOUND: {"2"}, Components.LOCAL: {"2"}})
    with uow:
        uow.entities.create_entity(create_identifier("1")).pull()
        uow.entities.create_entity(create_identifier("2")).delete()
        uow.commit()
    actual = get_entity_states(gateway, create_identifiers("1", "2"))
    expected = {(create_identifier("1"), states.Shared), (create_identifier("2"), states.Unshared)}
    assert actual == expected


def test_updates_are_discarded_on_context_exit() -> None:
    gateway, uow = initialize({Components.SOURCE: {"1", "2"}, Components.OUTBOUND: {"2"}, Components.LOCAL: {"2"}})
    with uow:
        uow.entities.create_entity(create_identifier("1")).pull()
        uow.entities.create_entity(create_identifier("2")).delete()
    actual = get_entity_states(gateway, create_identifiers("1", "2"))
    expected = {(create_identifier("1"), states.Unshared), (create_identifier("2"), states.Shared)}
    assert actual == expected


def test_updates_are_discarded_on_rollback() -> None:
    gateway, uow = initialize({Components.SOURCE: {"1", "2"}, Components.OUTBOUND: {"2"}, Components.LOCAL: {"2"}})
    with uow:
        uow.entities.create_entity(create_identifier("1")).pull()
        uow.entities.create_entity(create_identifier("2")).delete()
        uow.rollback()
    actual = get_entity_states(gateway, create_identifiers("1", "2"))
    expected = {(create_identifier("1"), states.Unshared), (create_identifier("2"), states.Shared)}
    assert actual == expected


def test_entities_can_not_be_accessed_outside_of_context() -> None:
    _, uow = initialize({Components.SOURCE: {"1"}})
    with uow:
        pass
    with pytest.raises(RuntimeError, match="outside"):
        uow.entities


def test_unable_to_commit_outside_of_context() -> None:
    _, uow = initialize({Components.SOURCE: {"1"}})
    with pytest.raises(RuntimeError, match="outside"):
        uow.commit()


def test_unable_to_rollback_outside_of_context() -> None:
    _, uow = initialize({Components.SOURCE: {"1"}})
    with pytest.raises(RuntimeError, match="outside"):
        uow.rollback()


def test_entity_expires_when_committing() -> None:
    _, uow = initialize({Components.SOURCE: {"1"}})
    with uow:
        entity = uow.entities.create_entity(create_identifier("1"))
        uow.commit()
        with pytest.raises(RuntimeError, match="expired entity"):
            entity.apply(Operations.START_PULL)


def test_entity_expires_when_rolling_back() -> None:
    _, uow = initialize({Components.SOURCE: {"1"}})
    with uow:
        entity = uow.entities.create_entity(create_identifier("1"))
        uow.rollback()
        with pytest.raises(RuntimeError, match="expired entity"):
            entity.apply(Operations.START_PULL)


def test_entity_expires_when_exiting_context() -> None:
    _, uow = initialize({Components.SOURCE: {"1"}})
    with uow:
        entity = uow.entities.create_entity(create_identifier("1"))
    with pytest.raises(RuntimeError, match="expired entity"):
        entity.apply(Operations.START_PULL)


def test_correct_events_are_collected() -> None:
    _, uow = initialize({Components.SOURCE: {"1", "2"}, Components.OUTBOUND: {"2"}, Components.LOCAL: {"2"}})
    with uow:
        uow.entities.create_entity(create_identifier("1")).pull()
        uow.entities.create_entity(create_identifier("2")).delete()
        uow.commit()
    expected = [
        events.StateChanged(
            Operations.START_PULL,
            create_identifier("1"),
            Transition(states.Unshared, states.Activated),
            Commands.START_PULL_PROCESS,
        ),
        events.StateChanged(
            Operations.PROCESS,
            create_identifier("1"),
            Transition(states.Activated, states.Received),
            Commands.ADD_TO_LOCAL,
        ),
        events.StateChanged(
            Operations.PROCESS,
            create_identifier("1"),
            Transition(states.Received, states.Shared),
            Commands.FINISH_PULL_PROCESS,
        ),
        events.StateChanged(
            Operations.START_DELETE,
            create_identifier("2"),
            Transition(states.Shared, states.Received),
            Commands.START_DELETE_PROCESS,
        ),
        events.StateChanged(
            Operations.PROCESS,
            create_identifier("2"),
            Transition(states.Received, states.Activated),
            Commands.REMOVE_FROM_LOCAL,
        ),
        events.StateChanged(
            Operations.PROCESS,
            create_identifier("2"),
            Transition(states.Activated, states.Unshared),
            Commands.FINISH_DELETE_PROCESS,
        ),
    ]
    actual = list(uow.collect_new_events())
    assert actual == expected


def test_unit_must_be_committed_to_collect_events() -> None:
    _, uow = initialize({Components.SOURCE: {"1"}})
    with uow:
        uow.entities.create_entity(create_identifier("1")).pull()
    assert list(uow.collect_new_events()) == []


def test_events_can_only_be_collected_once() -> None:
    _, uow = initialize({Components.SOURCE: {"1"}})
    with uow:
        uow.entities.create_entity(create_identifier("1")).pull()
        uow.commit()
    list(uow.collect_new_events())
    assert list(uow.collect_new_events()) == []


def test_events_can_only_be_collected_outside_of_context() -> None:
    _, uow = initialize({Components.SOURCE: {"1"}})
    with uow:
        uow.entities.create_entity(create_identifier("1")).pull()
        uow.commit()
        with pytest.raises(RuntimeError, match="inside context"):
            list(uow.collect_new_events())
